import pygame
from coopstructs.vectors import Vector2
from coopgame.colors import Color
from typing import Dict, Tuple

class MySprite(pygame.sprite.Sprite):
    def __init__(self, id:str, init_pos: Vector2, width: int, height: int):
        super().__init__()

        self.id = id

        # Pass in the color of the car, and its x and y position, width and height.
        # Set the background color and set it to be transparent
        self.surf = pygame.Surface([width, height]).convert()
        self.surf.fill(Color.PINK.value)
        self.surf.set_colorkey(Color.PINK.value)

        # Fetch the rectangle object that has the dimensions of the image.
        self.rect = self.surf.get_rect()
        self.set_pos(init_pos)

    def set_pos(self, pos: Vector2):
        self.rect = self.surf.get_rect(
            center=(
                pos.x,
                pos.y
            )
        )

    def blit(self, surface: pygame.Surface, display_handle: bool = False, display_rect: bool = False):

        surface.blit(self.surf, (self.rect.x, self.rect.y))
        if display_handle:
            pygame.draw.rect(surface, Color.RED.value, [self.rect.x, self.rect.y, 1, 1])
        if display_rect:
            pygame.draw.rect(surface, Color.RED.value, [self.rect.x, self.rect.y, self.rect.width, self.rect.height], 1)

class RectangleSprite(MySprite):
    def __init__(self, id:str, init_pos: Vector2, color: Color, width: int, height: int):
        # Call the parent class (Sprite) constructor
        MySprite.__init__(self, id, init_pos, width, height)

        pygame.draw.rect(self.surf, color.value, [0, 0, width, height])

class ImageSprite(MySprite):
    def __init__(self, id: str, init_pos: Vector2, color: Color, width: int, height: int):
        # Call the parent class (Sprite) constructor
        MySprite.__init__(self, id, init_pos, width, height)

class AnimatedSprite(MySprite):
    def __init__(self, id: str, init_pos: Vector2, animation_dict: Dict[str, Tuple], default_animation_key = None):
        self._animation_dict = animation_dict

        self._animation_key = default_animation_key
        self._animation_index = 0

        # Call the parent class (Sprite) constructor
        first_image = next(iter(self._animation_dict.values()))[0]
        MySprite.__init__(self, id, init_pos, first_image.get_width(), first_image.get_height())

        self.image = None
        self._set_image()

    def set_animation(self, animation: str):
        if animation in self._animation_dict.keys() and animation != self.current_animation:
            self._animation_key = animation
            self._animation_index = 0

    def increment_animation_phase(self, loop: bool = True):
        ''':return True if animation cycle has completed, False if still in cycle
        :param loop allows caller to specify whether or not to loop the aniimation (default True)'''
        ret = False
        self._animation_index += 1
        if self._animation_index >= len(self._animation_dict[self._animation_key]):
            self._animation_index = 0 if loop else len(self._animation_dict[self._animation_key]) - 1
            ret = True

        self._set_image()
        return ret

    def _set_image(self):
        self.surf.fill(Color.PINK.value)
        self.image = self._animation_dict[self._animation_key][self._animation_index]
        self.rect = self.image.get_rect( center = self.rect.center)
        self.surf.blit(self.image, (0, 0))

    @property
    def current_animation(self):
        return self._animation_key


from coopgame.renderedObjectHandling.wireframes.Wireframe import Wireframe
from typing import Tuple
import numpy as np
import logging
import MatrixManipulation as mm

class Cube(Wireframe):

    def __init__(self, origin: Tuple[int, int, int], size: int):
        super().__init__()
        self.size = size
        cube_nodes = [(x, y, z) for x in (origin[0], origin[0] + size) for y in (origin[1], origin[1] + size) for z in (origin[2], origin[2] + size)]
        self.addNodes(np.array(cube_nodes))
        self.addEdges(np.array([(n, n + 4) for n in range(0, 4)]))
        self.addEdges(np.array([(n, n + 1) for n in range(0, 8, 2)]))
        self.addEdges(np.array([(n, n + 2) for n in (0, 1, 4, 5)]))

class Triangle(Wireframe):

    def __init__(self, v1, v2, v3):
        super().__init__()
        self.addNodes(np.array([v1, v2, v3]))
        self.addEdges([(n, n + 1) for n in range(0, 2)])
        self.addEdges([(0, 2)])
        self.tuple = (v1, v2, v3)

class TriangleGridMesh(Wireframe):

    def __init__(self, grid_width: int, grid_height: int, scale: int):
        super().__init__()
        cols: int = int(grid_width / scale)
        rows: int = int(grid_height / scale)
        self.triangles = np.empty((0, 3), float)

        # self.triangles = []
        logging.info(f"Mesh being created: [{rows}] x [{cols}]")

        for ii in range(0, rows - 1):
            for jj in range(0, cols - 1):
                tri1 = Triangle((jj * scale, ii * scale, 0), ((jj + 1) * scale, ii * scale, 0),
                                    (jj * scale, (ii + 1) * scale, 0))
                tri2 =Triangle((jj * scale, (ii + 1) * scale, 0), ((jj + 1) * scale, (ii + 1) * scale, 0),
                                    ((jj + 1) * scale, ii * scale, 0))

                np.append(self.triangles, np.array(tri1.tuple), axis=0)
                np.append(self.triangles, np.array(tri2.tuple), axis=0)
                # self.triangles.append(tri1)
                # self.triangles.append(tri2)

                self.merge(tri1)
                self.merge(tri2)

    def update_mesh_heights(self, height_array):
        self.nodes[:, 2] = height_array
        # print (self.nodes)

if __name__ == "__main__":
    # cube = Cube((0,0,0), 1)
    # print(cube.edges)
    # cube.outputEdges()
    # cube.outputNodes()


    # triangle = Triangle((0, 0, 0), (25, 10, 50), (100, 125, 20))
    triangle = TriangleGridMesh(20, 20, 10)
    triangle.outputEdges()
    triangle.outputNodes()

    for ii in range(0, 10000):
        rotate = mm.rotateAroundPoint(triangle.findCentre(), [-.1, 0, 0])
        triangle.transform(rotate)

        triangle.outputNodes()

        rotate = mm.rotateAroundPoint(triangle.findCentre(), [.1, 0, 0])
        print(rotate)
        triangle.transform(rotate)

        triangle.outputNodes()


import numpy as np
from coopgame.coordinateConverter import CoordinateConverter as cc

class Wireframe:
    def __init__(self):
        self.nodes = np.zeros((0, 4), dtype=float)
        self.translation_coords = np.array([0, 0, 0], dtype=float)
        self.edge_node_ids = np.zeros((0, 2), dtype=int)
        self.object_centre = None

    def translated_nodes(self, matrix):
        return cc.apply_matrix_to_array(self.nodes, matrix)

    def translated_edges(self, matrix):
        t_nodes = self.translated_nodes(matrix=matrix)

        ret_edges = [(t_nodes[n1][:2], t_nodes[n2][:2]) for n1, n2 in self.edge_node_ids]

        return np.array(ret_edges)

    def addNodes(self, node_array):
        ones_column = np.ones((len(node_array), 1))
        ones_added = np.hstack((node_array, ones_column))
        self.nodes = np.vstack((self.nodes, ones_added))
        self.object_centre = self._calculate_object_centre()

    def addEdges(self, edgeList):
        self.edge_node_ids = np.concatenate((self.edge_node_ids, edgeList))
        # self.edges += edgeList

    def outputNodes(self):
        print ("\n --- Nodes --- ")
        for i, (x, y, z, _) in enumerate(self.nodes):
            print(" %d: (%.2f, %.2f, %.2f)" % (i, x, y, z))

    def outputEdges(self):
        print( "\n --- Edges --- ")
        for i, (node1, node2) in enumerate(self.edge_node_ids):
            print (f" {i}: {node1} -> {node2} [({self.nodes[node1][0]}, {self.nodes[node1][1]}) -> ({self.nodes[node2][0]}, {self.nodes[node2][1]})]" )

    def _calculate_object_centre(self):
        """ Find the centre of the wireframe. """
        num_nodes = len(self.nodes)
        meanX = sum([node[0] for node in self.nodes]) / num_nodes
        meanY = sum([node[1] for node in self.nodes]) / num_nodes
        meanZ = sum([node[2] for node in self.nodes]) / num_nodes
        return np.array((meanX, meanY, meanZ, 1))

    def findCentre(self, orientation_matrix):
        """ Find the centre of the wireframe. """
        return cc.apply_matrix_to_array(self.object_centre, orientation_matrix)

    def merge(self, wireframe):
        curLen = np.max([len(self.nodes), 0])
        self.addNodes(wireframe.nodes[:3, :-1])

        # logging.debug(f"Here are the existing edges: {self.edges}, CurLen: [{curLen}]")
        newEdges = np.zeros((0, 4), dtype=float) #[]
        # logging.debug(f"Here are the new edges: {wireframe.edges}")

        wireframe_edge_array = wireframe.edge_node_ids + curLen
        self.edge_node_ids = np.concatenate((self.edge_node_ids, wireframe_edge_array))
        # for edge in wireframe.edges:
        #     newEdge = np.array(edge[0] + curLen, edge[1] + curLen)
        #     newEdges = np.concatenate((newEdges, newEdge))

        # self.addEdges(newEdges)
        # logging.debug(self.edges)


if __name__ == "__main__":
    my_wireframe = Wireframe()
    my_wireframe.addNodes([(0, 0, 0), (1, 2, 3), (3, 2, 1)])
    my_wireframe.addEdges([(1, 2)])


from coopgame.pygridhandler import PyGridHandler
from coopgraph.grids import RectGrid
from coopgame.colors import Color
import pygame
import math
from typing import Hashable, Callable
from coopstructs.curves import LineCurve, Arc, Orientation
from coopstructs.geometry import Rectangle
from coopstructs.vectors import Vector2

class UnitPathHandler():

    def __init__(self, toggled_key: Hashable):
        self.toggled_key = toggled_key

    def generate(self
             , grid: RectGrid
             , grid_box_rect: Rectangle
             , naming_provider: Callable[[], str]):

        curves = []

        for row in range(0, grid.nRows):
            for col in range(0, grid.nColumns):

                cell = grid.at(row, col)
                if not cell.state[self.toggled_key].value:
                    continue

                left = cell.left.state[self.toggled_key].value if cell.left else None
                right = cell.right.state[self.toggled_key].value if cell.right else None
                up = cell.up.state[self.toggled_key].value if cell.up else None
                down = cell.down.state[self.toggled_key].value if cell.down else None
                connection_count = sum(x for x in [left, right, up, down] if x)

                if left and right:
                    # Horizontal Line
                    start = Vector2(col * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                    end = Vector2((col + 1) * grid_box_rect.width, (row + .5) * grid_box_rect.height)

                    curves.append(LineCurve(id = naming_provider(), origin=start, destination=end))

                if up and down:

                    # Vertical Line
                    start = Vector2((col + .5) * grid_box_rect.width, row * grid_box_rect.height)
                    end = Vector2((col + 0.5) * grid_box_rect.width, ((row + 1) * grid_box_rect.height))

                    curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                if up and right:
                    origin = Vector2((col + .5) * grid_box_rect.width, (row) * grid_box_rect.height)
                    curves.append(Arc(id = naming_provider(),
                                      orientation=Orientation.DOWN_RIGHT,
                                      origin=origin,
                                      arc_box_size=Vector2(grid_box_rect.width, grid_box_rect.height)))

                if up and left:
                    origin = Vector2((col) * grid_box_rect.width, (row + 0.5) * grid_box_rect.height)
                    curves.append(Arc(id=naming_provider(),
                                      orientation=Orientation.RIGHT_UP,
                                      origin=origin,
                                      arc_box_size=Vector2(grid_box_rect.width, grid_box_rect.height)))

                if right and down:
                    origin = Vector2((col + .5) * grid_box_rect.width, (row + 1) * grid_box_rect.height)
                    curves.append(Arc(id=naming_provider(),
                                      orientation=Orientation.UP_RIGHT,
                                      origin=origin,
                                      arc_box_size=Vector2(grid_box_rect.width, grid_box_rect.height)))


                if down and left:
                    origin = Vector2((col) * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                    curves.append(Arc(id=naming_provider(),
                                      orientation=Orientation.RIGHT_DOWN,
                                      origin=origin,
                                      arc_box_size=Vector2(grid_box_rect.width, grid_box_rect.height)))

                if connection_count == 1:
                    if up:
                        # Vertical Line
                        start = Vector2((col + .5) * grid_box_rect.width, row * grid_box_rect.height)
                        end = Vector2((col + 0.5) * grid_box_rect.width, ((row + 0.5) * grid_box_rect.height))

                        curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))
                    elif down:
                        # Vertical Line
                        start = Vector2((col + .5) * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                        end = Vector2((col + 0.5) * grid_box_rect.width, ((row + 1) * grid_box_rect.height))

                        curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                    elif right:
                        # Horizontal Line
                        start = Vector2((col + .5) * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                        end = Vector2((col + 1) * grid_box_rect.width, (row + .5) * grid_box_rect.height)

                        curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                    elif left:
                        # Horizontal Line
                        start = Vector2(col * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                        end = Vector2((col + .5) * grid_box_rect.width, (row + .5) * grid_box_rect.height)

                        curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                elif connection_count == 0:
                    # Horizontal Line
                    start = Vector2((col + 0.25) * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                    end = Vector2((col + .75) * grid_box_rect.width, (row + .5) * grid_box_rect.height)

                    curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                    # Vertical Line
                    start = Vector2((col + .5) * grid_box_rect.width, (row + .25) * grid_box_rect.height)
                    end = Vector2((col + .5) * grid_box_rect.width, (row + .75) * grid_box_rect.height)

                    curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

        return curves


from coopgame.colors import Color
import pygame
from coopgraph.graphs import Graph
from coopstructs.vectors import Vector2
from coopstructs.geometry import Rectangle
import coopgame.pygamehelpers as help
from typing import Callable, Tuple

class PyGraphHandler:
    def _draw_graph_edges(self,
                          surface: pygame.Surface,
                          graph: Graph,
                          coordinate_converter: Callable[[Vector2], Vector2] = None,
                          enabled_color: Color = None,
                          disabled_color: Color = None,
                          draw_scale_matrix = None,
                          draw_labels: bool = True,
                          label_offset: Tuple[int, int] = None,
                          label_font: pygame.Font = None
                          ):

        if not coordinate_converter:
            coordinate_converter = lambda x: x

        if not enabled_color:
            enabled_color = Color.BLUE

        if not disabled_color:
            disabled_color = Color.GREY

        if draw_labels:
            font = label_font if label_font is not None else pygame.font.Font(None, 25)
            label_offset = label_offset if label_offset is not None else (10, 10)

        for edge in graph.edges:
            start = coordinate_converter(edge.start.pos)
            end = coordinate_converter(edge.end.pos)

            width = 3 if graph.edge_between(edge.end, edge.start) is not None else 1

            scaled_pos = help.scaled_points_of_points([start, end], draw_scale_matrix=draw_scale_matrix)
            color = enabled_color if edge.enabled() else disabled_color
            pygame.draw.line(surface, color.value, (scaled_pos[0][0], scaled_pos[0][1]), (scaled_pos[1][0], scaled_pos[1][1]), width)

            if draw_labels:
                mid = ((scaled_pos[0][0] + scaled_pos[1][0]) / 2, (scaled_pos[0][1] + scaled_pos[1][1]) / 2 )
                help.draw_text(edge.id, surface, font, offset_rect=Rectangle(mid[0] + label_offset[0], mid[1] + label_offset[1], 100, 100))

    def _draw_graph_nodes(self,
                          surface: pygame.Surface,
                          graph: Graph,
                          coordinate_converter: Callable[[Vector2], Vector2] = None,
                          color: Color = None,
                          radius: int = 1,
                          width: int = 0,
                          draw_scale_matrix = None,
                          draw_labels: bool = True,
                          label_offset: Tuple[int, int] = None,
                          label_font: pygame.Font = None
                          ):
        if not coordinate_converter:
            coordinate_converter = lambda x: x

        if not color:
            color = Color.ORANGE

        if draw_labels:
            font = label_font if label_font is not None else pygame.font.Font(None, 25)
            label_offset = label_offset if label_offset is not None else (10, 10)

        for node in graph.nodes:
            position = coordinate_converter(node.pos)
            scaled_pos = help.scaled_points_of_points([position], draw_scale_matrix=draw_scale_matrix)
            position = (int(scaled_pos[0][0]), int(scaled_pos[0][1]))

            pygame.draw.circle(surface, color.value, position, radius, width)

            if draw_labels:
                help.draw_text(node.name, surface, font, offset_rect=Rectangle(scaled_pos[0][0] + label_offset[0], scaled_pos[0][1] + label_offset[1], 100, 100))

    @staticmethod
    def draw_articulation_points(surface: pygame.Surface,
                                 graph: Graph,
                                 color: Color = None
                                 ):
        if graph is None:
            return

        articulation_points = graph.AP()

        if color is None:
            color = Color.ORANGE

        for node, sub_graphs in articulation_points.items():
            pygame.draw.circle(surface, color.value, (int(node.pos.x), int(node.pos.y)), 10, 1)

    def draw_to_surface(self, surface: pygame.Surface,
                        graph: Graph,
                        coordinate_converter: Callable[[Vector2], Vector2] = None,
                        node_color: Color = None,
                        enabled_edge_color: Color = None,
                        disabled_edge_color: Color = None,
                        draw_scale_matrix = None):
        if not coordinate_converter:
            coordinate_converter = lambda x: x

        self._draw_graph_edges(surface, graph=graph, coordinate_converter=coordinate_converter, enabled_color=enabled_edge_color, disabled_color=disabled_edge_color, draw_scale_matrix=draw_scale_matrix)
        self._draw_graph_nodes(surface, graph=graph, coordinate_converter=coordinate_converter, color=node_color, draw_scale_matrix=draw_scale_matrix)



from pynput import keyboard
from typing import List, Tuple, Set
import logging
import threading

class InputState():
    def __init__(self):
        self.keyboard_state = {}
        self.keys_pressed_events = set()
        self.keys_released_events = set()

        self.last_state = {}
        self.current_state = {}

        self._lock = threading.Lock()
        listener = keyboard.Listener(
            on_press=self._on_press,
            on_release=self._on_release)
        listener.start()

    @property
    def pressed_keys(self):
        with self._lock:
            return [k for k, v in self.keyboard_state.items() if v == 1]

    def _on_press(self, key):
        try:
            key_val = str(key.char)
            logging.debug(('alphanumeric key {0} pressed'.format(key_val)))
        except AttributeError:
            key_val = str(key)
            logging.debug('special key {0} pressed'.format(key_val))

        with self._lock:
            if self.keyboard_state.get(key_val, 0) == 0:
                self.keys_pressed_events.add(key_val)
            self.keyboard_state[key_val] = 1


    def _on_release(self, key):
        with self._lock:
            try:
                key_val = str(key.char)
                logging.debug(('alphanumeric key {0} released'.format(key_val)))
            except AttributeError:
                key_val = str(key)
                logging.debug('special key {0} released'.format(key_val))

            self.keyboard_state[key_val] = 0
            self.keys_released_events.add(key_val)

    def is_pressed(self, keys):
        pressed = self.pressed_keys.copy()
        return self._check_if_all_are_pressed(pressed, keys)

    @staticmethod
    def _check_if_all_are_pressed(pressed, keys):
        if type(keys) in [List, tuple, set]:
            ret = all(x in pressed for x in keys)
        else:
            ret = keys in pressed
        logging.debug(f"Checking {keys} in {pressed} --> {ret}")
        return ret

    def clear_events(self):
        with self._lock:
            self.keys_pressed_events.clear()
            self.keys_released_events.clear()





import pygame
from coopstructs.geometry import Rectangle
from coopgame.colors import Color
from coopstructs.vectors import Vector2, IVector
import numpy as np
from typing import List


def mouse_pos_as_vector() -> Vector2:
    """ Get the global coords of the mouse position and convert them to a Vector2 object"""
    pos = pygame.mouse.get_pos()
    return Vector2(pos[0], pos[1])


def draw_box(surface: pygame.Surface, rect: Rectangle, color: Color, width: int = 0):
    pygame.draw.rect(surface, color.value, (rect.x, rect.y, rect.width, rect.height), width)


def draw_polygon(surface: pygame.Surface,
                 points,
                 color: Color,
                 width: int = 0,
                 alpha: int = 255):

    if type(points[0]) in [IVector, Vector2]:
        points = [point.as_tuple() for point in points]

    minx = min(point[0] for point in points)
    maxx = max(point[0] for point in points)
    miny = min(point[1] for point in points)
    maxy = max(point[1] for point in points)

    points = [(point[0] - minx, point[1] - miny) for point in points]

    # s = pygame.Surface((int(maxx - minx), int(maxy - miny)))
    # s.set_alpha(alpha)
    # s.set_colorkey(Color.HOT_PINK.value)
    # s.fill(Color.HOT_PINK.value)
    # pygame.draw.polygon(s, color.value, points, width)

    s = pygame.Surface((int(maxx - minx), int(maxy - miny)), pygame.SRCALPHA)
    c_with_alpha = (color.value[0], color.value[1], color.value[2], alpha)
    pygame.draw.polygon(s, c_with_alpha, points, width)

    surface.blit(s, (minx, miny))


def game_area_coords_from_parent_coords(parent_coords: Vector2, game_area_surface_rectangle: Rectangle) -> Vector2:
    """Converts Global Coords into coords on the game area"""
    return Vector2(parent_coords.x - game_area_surface_rectangle.x, parent_coords.y - game_area_surface_rectangle.y)


def scaled_points_of_a_rect(rect, grid_pos: Vector2, draw_scale_matrix=None, margin: int = 0):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    ''' get the rectangle object representing the grid position that was input'''
    rect = Rectangle(x=(margin + rect.width) * grid_pos.x + margin
                     , y=(margin + rect.height) * grid_pos.y + margin
                     , height=rect.height
                     , width=rect.width)

    '''Convert the rectangle to a list of points at the 4 corners'''
    points = [(x[0], x[1], 0, 1) for x in rect.points_tuple()]

    '''Multiply the points by the transform matrix for drawing'''
    transformed_points = draw_scale_matrix.dot(np.transpose(points))  # Transpose the points to appropriately mutiply

    '''return the x and y position for all points on the rectangle'''
    return np.transpose(transformed_points)[:,
           :3]  # Re-Transpose the points back to remain in a "list of points" format


def scaled_points_of_points(points: List[Vector2], draw_scale_matrix=None):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    if len(points) == 0:
        return np.array([])
    '''Convert the point to a 4-dim point for multiplication'''
    normal_points = [(point.x, point.y, 0, 1) for point in points]

    '''Multiply the points by the transform matrix for drawing'''
    transformed_points = draw_scale_matrix.dot(
        np.transpose(normal_points))  # Transpose the points to appropriately mutiply

    return np.transpose(transformed_points)[:,
           :3]  # Re-Transpose the points back to remain in a "list of points" format


def viewport_point_on_plane(viewport_point: Vector2, game_area_rect, draw_scale_matrix=None, margin: int = 0):
    points_on_plane = scaled_points_of_a_rect(game_area_rect, Vector2(0, 0), draw_scale_matrix, margin=margin)
    point0 = points_on_plane[0]
    point1 = points_on_plane[1]
    point2 = points_on_plane[2]

    vec1 = point0 - point1
    vec2 = point2 - point1

    normal = np.cross(vec1, vec2)
    a = normal[0]
    b = normal[1]
    c = normal[2]
    d = a * point0[0] + b * point0[1] + c * point0[2]

    z_val = (d - a * viewport_point.x - b * viewport_point.y) / c

    return [viewport_point.x, viewport_point.y, z_val]


def scaled_points_to_normal_points(points: List[IVector], draw_scale_matrix=None):
    translated_points = [(point.x, point.y, 0, 1) for point in points]
    normal_array = scaled_array_to_normal_array(scaled_array=translated_points, draw_scale_matrix=draw_scale_matrix)
    normal_points = [Vector2(point[0], point[1]) for point in normal_array]

    return normal_points


def scaled_array_to_normal_array(scaled_array, draw_scale_matrix=None):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    draw_scale_matrix_inv = np.linalg.inv(draw_scale_matrix)
    normal_points = draw_scale_matrix_inv.dot(np.transpose(scaled_array))

    normal_points = np.reshape(normal_points, (-1, 2))

    return normal_points


def normal_points_to_scaled_points(points: List[Vector2], draw_scale_matrix=None):
    return scaled_points_of_points(points, draw_scale_matrix)


def mouse_in_plane_point(game_area_rect: Rectangle, draw_scale_matrix=None):
    # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
    """Gets the mouse position and converts it to a grid position"""
    mouse_game_area_coord = game_area_coords_from_parent_coords(parent_coords=mouse_pos_as_vector(),
                                                                game_area_surface_rectangle=game_area_rect)
    mouse_plane_point = viewport_point_on_plane(mouse_game_area_coord, game_area_rect, draw_scale_matrix, margin=1)

    return mouse_plane_point


def normal_mouse_position(game_area_rect: Rectangle, draw_scale_matrix=None) -> Vector2:
    # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
    """Gets the mouse position and converts it to a grid position"""
    # mouse_game_area_coord = game_area_coords_from_parent_coords(parent_coords=mouse_pos_as_vector(), game_area_surface_rectangle=game_area_rect)
    mouse_plane_point = mouse_in_plane_point(game_area_rect, draw_scale_matrix)
    points = np.array([mouse_plane_point[0], mouse_plane_point[1], mouse_plane_point[2], 1])

    normal_array = scaled_array_to_normal_array(points, draw_scale_matrix)

    return Vector2(normal_array[0][0], normal_array[0][1])

def draw_text(text: str, hud: pygame.Surface, font: pygame.font.Font = None, offset_rect: Rectangle = None,
              color: Color = None):
    if font is None:
        font = pygame.font.Font(None, 30)

    if offset_rect is None:
        offset_rect = Rectangle(0, hud.get_height() - 50, 20, hud.get_width())

    if color is None:
        color = Color.BLUE

    rendered_txt = font.render(text, True, color.value)
    display1 = hud.subsurface(offset_rect.x, offset_rect.y, offset_rect.width, offset_rect.height)
    display1.blit(rendered_txt, display1.get_rect())


import pygame
import logging
from coopstructs.vectors import Vector2
from coopstructs.geometry import Rectangle
from coopgame.colors import Color
import functools
from coopgame.pygbutton import PygButton
from coopbugger.monitoredclass import MonitoredClass
from typing import Callable
import coopgame.pygamehelpers as help


def try_handler(func):
    @functools.wraps(func)
    def wrapper_handler(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except NotImplementedError as e:
            error = f"Inherited class should implement logic for {func.__name__}"
            logging.warning(error)
        except Exception as e:
            logging.exception(e)

    return wrapper_handler


class GameTemplate(MonitoredClass):

    def __init__(self, fullscreen: bool = False, screen_width: int = 1600, screen_height: int = 1000,
                 max_fps: int = 120, debug_mode=False):
        super().__init__()
        self.screen_width = screen_width
        self.screen_height = screen_height

        self.fullscreen = fullscreen
        self.screen = None
        self._init_screen(self.fullscreen)

        self.ticks = 0
        self.frame_times = []
        self.fps = None
        self.max_fps = max_fps
        self.clock = pygame.time.Clock()

        self.buttons = {}
        self._key_handlers = {
            (pygame.K_ESCAPE,): (self.quit, False),
            (pygame.K_F12,): (self.toggle_fullscreen, False),
            (pygame.K_F11,): (self.toggle_debug_mode, False)

        }
        self.register_keys()


        self.running = False
        self._debug_mode = debug_mode
        pygame.init()

    def quit(self):
        self.running = False

    def toggle_fullscreen(self):
        self.fullscreen = not self.fullscreen
        self._init_screen(self.fullscreen)

    def _init_screen(self, fullscreen):
        if fullscreen:
            self.screen = pygame.display.set_mode((0, 0), pygame.FULLSCREEN)
        else:
            self.screen = pygame.display.set_mode((self.screen_width, self.screen_height))

    def register_button(self, id, text, callback, postion_rect):
        self.buttons[id] = PygButton(postion_rect, caption=text, callback=callback)

    def main(self):

        self.initialize_game()

        self.running = True
        ii = 0
        while self.running:
            self._update()
            self._draw(frames=ii)
            self.clock.tick(self.max_fps)
            ii += 1

        pygame.quit()

    def calculate_fps(self, ticks_last_frame: int):
        if len(self.frame_times) > 20:
            self.frame_times.pop(0)

        self.frame_times.append(ticks_last_frame)

        avg_sec_per_frame = sum(self.frame_times) / len(self.frame_times) / 1000.0
        self.fps = 1 / avg_sec_per_frame if avg_sec_per_frame > 0 else 0

    def _update(self):
        """:return
            Update environment based on time delta and any input
        """

        ''' Calculate the ticks between update calls so that the update functions can handle correct time deltas '''
        t = pygame.time.get_ticks()
        delta_time_ms = (t - self.ticks)
        self.ticks = t
        self.calculate_fps(delta_time_ms)

        '''Update Model'''
        self.model_updater(delta_time_ms)

        '''Update Sprites'''
        self.sprite_updater(delta_time_ms)

        '''Handle Events'''
        self._handle_events()

    def _handle_events(self):
        """:return
            handle all of the registered events that have been captured since last iteration
        """

        '''Get next event'''
        for event in pygame.event.get():

            '''Get pressed keys'''
            pressed_keys = pygame.key.get_pressed()

            '''Check and handle button press'''
            self.handle_buttons(event)

            '''Debug Printer'''
            if event.type not in (0, 1, 4, 6):
                logging.debug(f"Pygame EventType: {event.type}")

            '''Event Type Switch'''
            if event.type == pygame.QUIT:
                self.quit()
            elif event.type == pygame.KEYDOWN:
                self._handle_key_pressed(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
                logging.info(f"Left click")
                self.handle_left_click(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 3:
                logging.info(f"Right click")
                self.handle_right_click(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 4:
                logging.info(f"Scroll Up")
                self.handle_mouse_scroll_up(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 5:
                logging.info(f"Scroll Down")
                self.handle_mouse_scroll_down(pressed_keys)
            elif event.type == pygame.VIDEORESIZE:
                logging.info(f"Window Resized [{self.screen.get_width()}x{self.screen.get_height()}]")
                self.on_resize()

        '''Get pressed keys'''
        pressed_keys = pygame.key.get_pressed()

        '''Handle Held Keys'''
        if pressed_keys is not None:
            self._handle_held_keys(pressed_keys)

        '''Handle hover over'''
        self._handle_hover_over()

    def register_action_to_keys(self, keys_tuple, func: Callable, react_while_holding: bool = False):
        """
            Takes a tuple of keys (integers) and maps that key combo to a callable function

            :param keys_tuple a list of keys (integers) that will be mapped to the input callable. Note that a single
            value Tuple is input as ([key],) *note the comma
            :param func a callable that is mapped to the key combo
        """
        self._key_handlers[keys_tuple] = (func, react_while_holding)

    @MonitoredClass.timer
    @try_handler
    def handle_buttons(self, event):
        for id, button in self.buttons:
            if 'click' in button.handleEvent(event):
                button.callback()

    @MonitoredClass.timer
    @try_handler
    def _handle_key_pressed(self, pressed_keys):
        buttons = [pygame.key.name(k) for k, v in enumerate(pressed_keys) if v]
        logging.info(f"Keys pressed: {buttons}")
        for mapped_keys, reaction in self._key_handlers.items():
            func = reaction[0]
            if (all(pressed_keys[key] for key in mapped_keys)):
                func()

    @MonitoredClass.timer
    @try_handler
    def _handle_held_keys(self, pressed_keys):
        for mapped_keys, reaction in self._key_handlers.items():
            func = reaction[0]
            react_while_holding = reaction[1]
            if (all(pressed_keys[key] for key in mapped_keys)) and react_while_holding:
                func()

    @MonitoredClass.timer
    def _handle_hover_over(self):
        mouse_pos_as_vector = help.mouse_pos_as_vector()
        if mouse_pos_as_vector:
            self.handle_hover_over(mouse_pos_as_vector)

    @MonitoredClass.timer
    @try_handler
    def initialize_game(self):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_left_click(self, pressed_keys):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_right_click(self, pressed_keys):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_mouse_scroll_up(self, pressed_keys):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_mouse_scroll_down(self, pressed_keys):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def draw(self, frames: int, debug_mode: bool = False):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def model_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def sprite_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def on_resize(self):
        raise NotImplementedError()

    @try_handler
    def register_keys(self):
        raise NotImplementedError()

    @MonitoredClass.timer
    def _draw(self, frames: int):
        self.screen.fill(Color.BLACK.value)
        self.draw(frames, self._debug_mode)
        # Update the display
        pygame.display.flip()

    def draw_mouse_coord(self, hud: pygame.Surface, font: pygame.font.Font = None, offset_rect: Rectangle = None,
                         color: Color = None):
        mouse_pos_as_vector = help.mouse_pos_as_vector()
        txt = f"M:<{int(mouse_pos_as_vector.x)}, {int(mouse_pos_as_vector.y)}>"
        self.draw_text(txt, hud=hud, font=font, offset_rect=offset_rect, color=color)

    def draw_fps(self, hud: pygame.Surface, font: pygame.font.Font = None, offset_rect: Rectangle = None,
                 color: Color = None):
        txt = f"FPS: {int(self.fps)}"
        self.draw_text(txt, hud=hud, font=font, offset_rect=offset_rect, color=color)

    def draw_text(self, text: str, hud: pygame.Surface, font: pygame.font.Font = None, offset_rect: Rectangle = None,
                  color: Color = None):
        if font is None:
            font = pygame.font.Font(None, 30)

        if offset_rect is None:
            offset_rect = Rectangle(0, hud.get_height() - 50, 20, hud.get_width())

        if color is None:
            color = Color.BLUE

        rendered_txt = font.render(text, True, color.value)
        display1 = hud.subsurface(offset_rect.x, offset_rect.y, offset_rect.width, offset_rect.height)
        display1.blit(rendered_txt, display1.get_rect())

    def window_rect(self):
        return Rectangle(0, 0, self.screen.get_height(), self.screen.get_width())

    def toggle_debug_mode(self):
        self._debug_mode = not self._debug_mode

    def draw_monitoredclass_stats(self, monitoredClass: MonitoredClass, surface: pygame.Surface,
                                  font: pygame.font.SysFont = None, offset_rectangle: Rectangle = None):
        if font is None:
            font = pygame.font.SysFont(None, 18)

        if offset_rectangle is None:
            offset_rectangle = Rectangle(0, 0, font.get_height() + 3, 300)

        tracked_time = [(key, val) for key, val in monitoredClass.tracked_time.items()]
        tracked_time.sort(key=lambda x: x[1], reverse=True)

        y_off = offset_rectangle.y
        offset_rectangle.y = y_off
        self.draw_text(f"RunTime Stats for {type(monitoredClass).__name__}", surface, color=Color.LEMON_CHIFFON,
                       offset_rect=offset_rectangle, font=font)
        y_off += font.get_height() + 3
        for key, val in tracked_time:
            offset_rectangle.y = y_off
            self.draw_text(f"{key}: {round(val, 2)} sec", surface, color=Color.LEMON_CHIFFON,
                           offset_rect=offset_rectangle, font=font)
            y_off += font.get_height() + 3


import pygame
from coopgame.colors import Color
from typing import List, Dict, Callable
from coopstructs.curves import Arc, CubicBezier, LineCurve, Curve, Orientation, CatmullRom, CircularArc
from coopstructs.geometry import Line
import numpy as np


class CurveDrawer:

    def __init__(self):
        pass

    def draw_curves(self,
                    curves: Dict[Curve, Color],
                    surface: pygame.Surface,
                    draw_control_points: bool = True,
                    draw_control_lines: bool = True,
                    control_point_color: Color = Color.RED,
                    control_point_size: int = None,
                    control_line_color: Color = Color.GREY,
                    draw_scale_matrix=None):
        for curve, color in curves.items():
            if type(curve) == CubicBezier:
                self.draw_bezier(surface,
                                 curve,
                                 color,
                                 draw_points=draw_control_points,
                                 control_point_color=control_point_color,
                                 control_point_size=control_point_size,
                                 draw_control_lines=draw_control_lines,
                                 control_line_color=control_line_color,
                                 draw_scale_matrix=draw_scale_matrix)
            elif type(curve) == LineCurve:
                self.draw_line(surface,
                               curve,
                               color,
                               draw_points=draw_control_points,
                               control_point_color=control_point_color,
                               control_point_size=control_point_size,
                               draw_scale_matrix=draw_scale_matrix)
            elif type(curve) == Arc:
                self.draw_arc(surface,
                              curve,
                              color,
                              draw_points=draw_control_points,
                              control_point_color=control_point_color,
                              control_point_size=control_point_size,
                              draw_scale_matrix=draw_scale_matrix)
            elif type(curve) == CatmullRom:
                self.draw_catmullrom(surface,
                                     curve,
                                     color,
                                     draw_points=draw_control_points,
                                     control_point_color=control_point_color,
                                     control_point_size=control_point_size,
                                     draw_scale_matrix=draw_scale_matrix)
            elif type(curve) == CircularArc:
                self.draw_circulararc(surface,
                                      curve,
                                      color,
                                      draw_points=draw_control_points,
                                      control_point_color=control_point_color,
                                      control_point_size=control_point_size,
                                      draw_scale_matrix=draw_scale_matrix)
            else:
                raise NotImplementedError(f"Unahanlded curve type {curve}")

    def draw_bezier(self,
                    screen,
                    curve: CubicBezier,
                    curve_color: Color,
                    draw_points: bool = False,
                    control_point_color: Color = Color.RED,
                    control_point_size: int = None,
                    draw_control_lines: bool = False,
                    control_line_color: Color = Color.GREY,
                    draw_scale_matrix=None):

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert control points into np.arrays
        arrays_of_points = []
        for point in curve.control_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        # Draw control points
        if control_point_size is None:
            control_point_size = 4

        if draw_points:
            for ii in range(len(translated_points)):
                pygame.draw.circle(screen, control_point_color.value,
                                   (int(translated_points[ii][0]), int(translated_points[ii][1])), control_point_size)
            # for p in curve.control_points:
            #     pygame.draw.circle(screen, control_point_color.value, (int(p.x), int(p.y)), 4)

        # Draw control "lines"
        if draw_control_lines:
            pygame.draw.lines(screen,
                              control_line_color.value,
                              False,
                              [(translated_points[ii][0], translated_points[ii][1])
                               for ii in range(len(translated_points))])
            # pygame.draw.lines(screen, control_line_color.value, False, [(x.x, x.y) for x in curve.control_points])

        # Draw bezier curve
        b_points = curve.compute_bezier_points([(translated_points[ii][0], translated_points[ii][1])
                                                for ii in range(len(translated_points))])
        # b_points = curve.compute_bezier_points([(x.x, x.y) for x in curve.control_points])
        if b_points is not None and len(b_points) > 1:
            pygame.draw.lines(screen, curve_color.value, False, b_points, 2)

    def draw_line(self,
                  surface: pygame.Surface,
                  line: Line,
                  color: Color,
                  draw_points: bool = False,
                  control_point_color: Color = Color.RED,
                  control_point_size: int = None,
                  draw_scale_matrix=None):
        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert Origin/Destination into np.arrays
        origin = np.array([line.origin.x, line.origin.y, 0, 1])
        destination = np.array([line.destination.x, line.destination.y, 0, 1])

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array([origin, destination]))).transpose()

        # Draw the lines
        pygame.draw.line(surface, color.value, (translated_points[0][0], translated_points[0][1]),
                         (translated_points[1][0], translated_points[1][1]))

        # Draw the points
        if control_point_size is None:
            control_point_size = 4

        if draw_points:
            pygame.draw.circle(surface, control_point_color.value,
                               (int(translated_points[0][0]), int(translated_points[0][1])), control_point_size)
            pygame.draw.circle(surface, control_point_color.value,
                               (int(translated_points[1][0]), int(translated_points[1][1])), control_point_size)

    def draw_arc(self,
                 surface: pygame.Surface,
                 arc: Arc,
                 color: Color,
                 draw_points: bool = False,
                 control_point_color: Color = Color.RED,
                 control_point_size: int = None,
                 draw_scale_matrix=None):

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert control points into np.arrays
        control_points = arc.compute_curve_points()
        arrays_of_points = []
        for point in control_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        # Draw Lines
        pygame.draw.lines(surface, color.value, False, [(translated_points[ii][0], translated_points[ii][1])
                                                        for ii in range(len(translated_points))], 2)

        # Draw the points
        if control_point_size is None:
            control_point_size = 4

        if draw_points:
            pygame.draw.circle(surface, control_point_color.value,
                               (int(translated_points[0][0]), int(translated_points[0][1])), control_point_size)
            pygame.draw.circle(surface, control_point_color.value,
                               (int(translated_points[-1][0]), int(translated_points[-1][1])), control_point_size)

    def draw_catmullrom(self,
                        screen,
                        curve: CatmullRom,
                        curve_color: Color,
                        draw_points: bool = False,
                        control_point_color: Color = Color.RED,
                        control_point_size: int = None,
                        draw_control_lines: bool = False,
                        control_line_color: Color = Color.GREY,
                        draw_scale_matrix=None):

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert control points into np.arrays
        arrays_of_points = []
        for point in curve.control_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        # Draw control points
        if control_point_size is None:
            control_point_size = 4

        if draw_points:
            for ii in range(len(translated_points)):
                pygame.draw.circle(screen, control_point_color.value,
                                   (int(translated_points[ii][0]), int(translated_points[ii][1])), control_point_size)

        # Draw control "lines"
        if draw_control_lines:
            pygame.draw.lines(screen,
                              control_line_color.value,
                              False,
                              [(translated_points[ii][0], translated_points[ii][1])
                               for ii in range(len(translated_points))])

        # Draw CatmullRom curve
        b_points = curve.compute_catmull_points([(translated_points[ii][0], translated_points[ii][1])
                                                 for ii in range(len(translated_points))])
        # b_points = curve.compute_catmull_points(curve.control_points)

        if b_points is not None and len(b_points) > 1:
            pygame.draw.lines(screen, curve_color.value, False, [(point.x, point.y) for point in b_points], 2)

    def draw_circulararc(self,
                         screen,
                         curve: CircularArc,
                         curve_color: Color,
                         draw_points: bool = False,
                         control_point_color: Color = Color.RED,
                         control_point_size: int = None,
                         draw_control_lines: bool = False,
                         control_line_color: Color = Color.GREY,
                         draw_scale_matrix=None):
        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert control points into np.arrays
        arrays_of_points = []
        for point in curve.control_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        # Draw control points
        if control_point_size is None:
            control_point_size = 4

        if draw_points:
            for ii in range(len(translated_points)):
                pygame.draw.circle(screen, control_point_color.value,
                                   (int(translated_points[ii][0]), int(translated_points[ii][1])), control_point_size)

        # Draw control "lines"
        if draw_control_lines:
            pygame.draw.lines(screen,
                              control_line_color.value,
                              False,
                              [(translated_points[ii][0], translated_points[ii][1])
                               for ii in range(len(translated_points))])

        # Draw circulararc_points
        b_points = curve.compute_circulararc_points()

        # Convert
        arrays_of_points = []
        for point in b_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        if translated_points is not None and len(translated_points) > 1:
            pygame.draw.lines(screen, curve_color.value, False, [(point[0], point[1]) for point in translated_points],
                              2)

    def draw_curve_builder(self):
        pass